from typing import Callable, List
from eth_vertigo.core.filter import MutationFilter

from eth_vertigo.core import Mutation, MutationResult, TestSuggester
from jinja2 import PackageLoader, Environment
from concurrent.futures import ThreadPoolExecutor


environment = Environment(
        loader=PackageLoader("eth_vertigo.core"), trim_blocks=True
    )


class CampaignReport:
    def __init__(self, mutations):
        """
        Constructs a campaign report object
        :param mutations: Mutations for this campaign
        """
        self.mutations = mutations

    @property
    def mutation_count(self):
        """ Amount of mutations in this report """
        return len(self.mutations)

    @property
    def nr_killed(self):
        """ Amount of killed mutations in this report"""
        return len([e for e in self.mutations if e.result == MutationResult.KILLED])

    @property
    def nr_alive(self):
        """ Amount of alive mutations in this report"""
        return len([e for e in self.mutations if e.result == MutationResult.LIVED])

    def render(self, with_mutations=False):
        template = environment.get_template("report_template.jinja2")
        return template.render(
            nr_mutations=len(self.mutations),
            nr_killed=self.nr_killed,
            mutations=self.mutations,
            with_mutations=with_mutations
        )


class Campaign:
    """
    A core campaign class orchestrates and manages a core testing run
    """

    def __init__(self, filters: List[MutationFilter] = None , suggesters=None):
        self.sources = []
        self.project_directory = None
        self.mutations = []
        self.is_set_up = False
        self.filters = filters or []
        self.suggesters = suggesters or []  # type: List[TestSuggester]

    def setup(self):
        """ Sets up the campaign for execution"""
        raise NotImplementedError

    def run(self, progress_callback: Callable, threads=1):
        """ Starts a core testing campaign"""
        if not self.is_set_up:
            raise ValueError("This campaign is not setup yet")

        report = CampaignReport(self.mutations)
        with ThreadPoolExecutor(max_workers=threads) as e:
            for mutation in self.mutations:
                e.submit(self.test_mutation, mutation, progress_callback)
            e.shutdown()
        return report

    def test_mutation(self, mutation: Mutation, done_callback: Callable):
        """ Run the test suite using a core and check for murders """
        raise NotImplementedError

    def valid(self):
        """ Checks whether the current project is valid """
        raise NotImplementedError

    def store_compilation_results(self):
        """ Stores compilation results for trivial compiler equivalence"""
        raise NotImplementedError



class Runner:

    @property
    def tests(self) -> list:
        """ Returns a list of the tests """
        raise NotImplementedError

    def run_tests(self, coverage: bool):
        """ Runs all the tests that are available"""
        raise NotImplementedError

    def run_test(self, name: str, coverage: bool):
        """ Runs the test with the given name """
        raise NotImplementedError



from subprocess import Popen, run, getoutput,  PIPE
from typing import Optional
from tempfile import TemporaryFile
from time import sleep
from loguru import logger
DEFAULT_GANACHE_PARAMETERS = []  # ["--dbMemdown"]


class Ganache:
    def __init__(self, port, parameters, ganache_binary="ganache"):
        # Remove any pre-set port options
        self.parameters = parameters
        self.parameters.extend(["--port", str(port)])

        for param in DEFAULT_GANACHE_PARAMETERS:
            if param in self.parameters:
                continue
            self.parameters.append(param)

        self.ganache_binary = ganache_binary
        self.process = None  # type: Optional[subprocess.Popen]

    def start(self):
        if self.process is not None:
            raise ValueError("Process has already been terminated")

        self.process = Popen(
            [self.ganache_binary] + self.parameters,
            stderr=PIPE, stdout=PIPE
        )

        while True:
            line = self.process.stdout.readline()
            if "Listening on" in str(line):
                break

        if self.process.poll() is not None:
            raise Exception("Could not create ganache network")

    def stop(self):
        if self.process is None:
            raise ValueError("Process has not yet been started")
        if self.process.poll():
            raise ValueError("Process has already terminated")
        self.process.terminate()



import click
from os import getcwd
from pathlib import Path
from eth_vertigo.core import MutationResult
from eth_vertigo.core.network import DynamicNetworkPool, StaticNetworkPool, Ganache
from eth_vertigo.core.truffle.truffle_campaign import TruffleCampaign
from eth_vertigo.core.filters.sample_filter import SampleFilter
from eth_vertigo.core.filters.exclude_filter import ExcludeFilter
from eth_vertigo.test_runner.truffle import TruffleRunnerFactory
from eth_vertigo.test_runner.exceptions import TestRunException
from eth_vertigo.interfaces.truffle import Truffle
from eth_vertigo.mutator.universal_mutator import UniversalMutator

from eth_vertigo.incremental import IncrementalRecorder, IncrementalMutationStore, IncrementalSuggester

from tqdm import tqdm


@click.group(help="Mutation testing framework for smart contracts")
def cli():
    pass


@cli.command(help="Performs a core test campaign")
@click.option('--output', help="Output core test results to file", nargs=1, type=str)
@click.option('--network', help="Network names that vertigo can use", multiple=True)
@click.option('--ganache-path', help="Path to ganache binary", type=str, default="ganache-cli")
@click.option('--ganache-network', help="Dynamic networks that vertigo can use eg. (develop, 8485)",
              multiple=True, type=(str, int))
@click.option('--ganache-network-options', help="Options to pass to dynamic ganache networks", type=str)
@click.option('--rules', help="Universal Mutator style rules to use in mutation testing", multiple=True)
@click.option('--truffle-location', help="Location of truffle cli", nargs=1, type=str, default="truffle")
@click.option('--sample-ratio', help="If this option is set. Vertigo will apply the sample filter with the given ratio", nargs=1, type=float)
@click.option('--exclude', help="Vertigo won't mutate files in these directories", multiple=True)
@click.option('--incremental', help="File where incremental mutation state is stored",
              type=str)
def run(
        output,
        network,
        ganache_path,
        ganache_network,
        ganache_network_options,
        rules,
        truffle_location,
        sample_ratio,
        exclude,
        incremental
):
    """ Run command """
    click.echo("[*] Starting mutation testing")

    # Setup global parameters
    truffle = Truffle(truffle_location)

    working_directory = getcwd()
    project_type = _directory_type(working_directory)
    filters = []
    if exclude:
        filters.append(ExcludeFilter(exclude))

    if network and ganache_network:
        click.echo("[-] Can't use dynamic networks and regular networks simultaniously")
        exit(1)

    test_suggesters = []
    if incremental:
        incremental_store_file = Path(incremental)
        if not incremental_store_file.exists():
            pass
        elif not incremental_store_file.is_file() :
            click.echo(f"Incremental file {incremental} is not a file")
            exit(0)
        else:
            store = IncrementalMutationStore.from_file(incremental_store_file)
            test_suggesters.append(IncrementalSuggester(store))

    if project_type == "truffle":
        click.echo("[*] Starting analysis on truffle project")
        project_path = Path(working_directory)

        if not (project_path / "contracts").exists():
            click.echo("[-] No contracts directory in truffle project")
            return
        elif not (project_path / "test").exists():
            click.echo("[-] No test directory found in truffle project")
            return

        if sample_ratio:
            filters.append(SampleFilter(sample_ratio))

        mutators = []
        if rules:
            um = UniversalMutator()
            for rule_file in rules:
                um.load_rule(Path(rule_file))
            mutators.append(um)

        if network:
            network_pool = StaticNetworkPool(network)
        elif ganache_network:
            network_pool = DynamicNetworkPool(
                ganache_network,
                lambda port: Ganache(port, ganache_network_options.split(' ') if ganache_network_options else [],
                                     ganache_path)
            )
        else:
            click.echo("[-] Vertigo needs at least one network to run analyses on")
            return

        try:
            campaign = TruffleCampaign(
                project_directory=project_path,
                truffle_compiler=truffle,
                mutators=mutators,
                truffle_runner_factory=TruffleRunnerFactory(truffle),
                network_pool=network_pool,
                filters=filters,
                suggesters=test_suggesters,
            )
        except:
            click.echo("[-] Encountered an error while setting up the core campaign")
            for node in network_pool.claimed_networks.keys():
                click.echo(f"[+] Cleaning up network: {node}")
                network_pool.yield_network(node)
            raise
    else:
        click.echo("[*] Could not find supported project directory in {}".format(working_directory))
        return

    click.echo("[*] Initializing campaign run ")

    try:
        campaign.setup()
        click.echo("[*] Checking validity of project")
        if not campaign.valid():
            click.echo("[-] We couldn't get valid results by running the truffle tests.\n Aborting")
            return

        click.echo("[+] The project is valid")
        click.echo("[*] Storing compilation results")
        campaign.store_compilation_results()
        click.echo("[*] Running analysis on {} mutants".format(len(campaign.mutations)))
        with tqdm(total=len(campaign.mutations), unit="mutant") as pbar:
            report = campaign.run(lambda: pbar.update(1) and pbar.refresh(), threads=max(network_pool.size, 1))
        pbar.close()

    except TestRunException as e:
        click.echo("[-] Encountered an error while running the framework's test command:")
        click.echo(e)
        return
    except Exception as e:
        click.echo("[-] Encountered an error while running the core campaign")
        click.echo(e)
        raise

    click.echo("[*] Done with campaign run")
    click.echo("[+] Report:")
    click.echo(report.render())

    click.echo("[+] Survivors")
    for mutation in report.mutations:
        if mutation.result == MutationResult.LIVED:
            click.echo(str(mutation))

    if output:
        output_path = Path(output)
        if not output_path.exists() or click.confirm("[*] There already exists something at {}. Overwrite ".format(str(output_path))):
            click.echo("Result of core run can be found at: {}".format(output))
            output_path.write_text(report.render(with_mutations=True), "utf-8")

    if incremental:
        incremental_store_file = Path(incremental)
        if not incremental_store_file.exists() or \
            click.confirm(f"[*] There already exists an incremental at {incremental_store_file.name}. Overwrite "):
            new_incremental_store = IncrementalRecorder().record(report.mutations)
            new_incremental_store.to_file(incremental_store_file)

    click.echo("[*] Done! ")


def _directory_type(working_directory: str):
    """ Determines the current framework in the current directory """
    # jk, we only deal with truffle for now ^^'
    return "truffle"



import yaml
from typing import List, Dict
from pathlib import Path


class IncrementalMutationStore:
    def __init__(self):
        self.known_mutations = []  # type: List[MutationRecord]

    @property
    def yaml(self):
        return yaml.dump({"known_mutations": [m.as_dict for m in self.known_mutations]})

    @staticmethod
    def from_yaml(data):
        values = yaml.safe_load(data)
        result = IncrementalMutationStore()
        result.known_mutations = [MutationRecord.from_dict(record) for record in values.get("known_mutations", [])]
        return result

    @staticmethod
    def from_file(file: Path):
        if not file.is_file():
            raise ValueError("Passed path is not a file")

        content = file.read_text('utf-8')
        return IncrementalMutationStore.from_yaml(content)

    def to_file(self, file: Path):
        if file.exists() and not file.is_file():
            raise ValueError("Passed path already exists and is not a file")

        file.write_text(self.yaml, "utf-8")


class MutationRecord:
    def __init__(self):
        self.location = None
        self.original_text = None
        self.line_number = None
        self.source_file_name = None
        self.new_text = None
        self.crime_scenes = []  # type: List[str]

    @staticmethod
    def from_dict(data: Dict):
        result = MutationRecord()
        result.location = data.get("location", "")
        result.original_text = data.get("original_text", "")
        result.line_number = data.get("line_number", "")
        result.source_file_name = data.get("source_file_name", "")
        result.new_text = data.get("new_text", "")
        result.crime_scenes = data.get("crime_scenes", [])
        return result

    @property
    def as_dict(self):
        return self.__dict__




from eth_vertigo.test_runner.truffle.truffle_runner import TruffleRunner
from eth_vertigo.test_runner.truffle.truffle_runner_factory import TruffleRunnerFactory
from eth_vertigo.test_runner.truffle.truffle_tester import TruffleTester


from typing import List

from eth_vertigo.incremental.store import MutationRecord, IncrementalMutationStore
from eth_vertigo.core import Mutation


class IncrementalRecorder:
    def record(self, mutations: List[Mutation]) -> IncrementalMutationStore:
        store = IncrementalMutationStore()
        store.known_mutations = list(
            map(
                self._mutation_to_record,
                [m for m in mutations if m.crime_scenes]
            )
        )
        return store

    @staticmethod
    def _mutation_to_record(mutation: Mutation) -> MutationRecord:
        result = MutationRecord()
        result.new_text = mutation.value
        result.original_text = mutation.original_value
        result.source_file_name = mutation.source_file_name
        result.location = ":".join(map(str, mutation.location))
        result.line_number = mutation.line_number
        result.crime_scenes = mutation.crime_scenes
        return result



from pathlib import Path
from typing import List

from eth_vertigo.mutator.universal_mutator.rule import Rule
from eth_vertigo.mutator.universal_mutator.rule_loader import RuleLoader
from eth_vertigo.mutator.mutator import Mutator
from eth_vertigo.mutator.source_file import SourceFile
from eth_vertigo.core.mutation import Mutation


class UniversalMutator(Mutator):
    """Mutator  based on Universal Mutator

    Props to the design of universal mutator style mutation rules and evaluation of the approach go to Alex Groce et al.
    Repository:
        https://github.com/agroce/universalmutator
    Paper:
        An Extensible, Regular-Expression-Based Tool for Multi-Language Mutant Generation - Alex Groce et al.
    """

    def __init__(self):
        self.rule_sets = {}

    def load_rule(self, rule_file: Path):
        """Load rule from rule files"""
        loader = RuleLoader()
        rules = list(loader.load_from_file(rule_file))
        self.rule_sets[str(rule_file.name)] = rules

    def mutate(self, source_file: SourceFile, project_directory: Path) -> List[Mutation]:
        """Generate mutants

        Generates mutants based on loaded universal mutator rules
        """
        mutants = []
        for rule_file, rules in self.rule_sets.items():
            for rule in rules:
                mutants.extend(list(rule.generate_mutants(source_file, project_directory)))
        return mutants



from eth_vertigo.test_runner.truffle.truffle_tester import TruffleTester
from eth_vertigo.test_runner.truffle.truffle_runner import TruffleRunner


class TruffleRunnerFactory:
    """ Factory for TestRunner objects """
    def __init__(self, truffle_tester: TruffleTester) -> None:
        """Initializes a truffle runner factory"""
        self.truffle_tester = truffle_tester

    def create(self, project_directory: str) -> TruffleRunner:
        """ Create a new TruffleRunner

        :param project_directory: Project directory that the test runner should be aimed at
        :return: the brand new truffle runner
        """
        return TruffleRunner(
            project_directory=project_directory,
            truffle_tester=self.truffle_tester
        )



from typing import Dict


class TruffleCompiler:
    """Truffle compiler interface exposes compiling functionality from truffle"""
    def run_compile_command(self, working_directory: str) -> None:
        """ Runs truffle's test command

        :param working_directory: The truffle project directory to compile
        """
        raise NotImplementedError

    def get_bytecodes(self, working_directory: str) -> Dict[str, str]:
        """ Returns the bytecodes in the compilation result of the current directory

        :param working_directory: The truffle directory for which we retreive the bytecodes
        :return: bytecodes in the shape {'contractName': '0x00'}
        """
        raise NotImplementedError



from eth_vertigo.core import Mutation, TestSuggester

from typing import List, Tuple
from eth_vertigo.incremental.store import IncrementalMutationStore, MutationRecord


class IncrementalSuggester(TestSuggester):
    def __init__(self, store: IncrementalMutationStore):
        self._store = store

    @property
    def is_strict(self) -> bool:
        return False

    def _equivalent(self, mutation: Mutation, record: MutationRecord):
        return record.source_file_name == mutation.source_file_name \
               and record.original_text == mutation.original_value \
               and record.new_text == mutation.value \
               and record.line_number == mutation.line_number

    def suggest_tests(self, mutation: Mutation) -> List:
        records = self._store.known_mutations
        relevant_records = [r for r in records if self._equivalent(mutation, r)]
        tests = []
        for r in relevant_records:
            tests.extend(r.crime_scenes)
        return list(set(tests))





from abc import abstractmethod, ABC
from typing import Tuple, Callable, List, Dict


class Network:
    def __init__(self, name: str, port: int, provider=None):
        self.name = name
        self.port = port
        self.provider = provider


class NetworkPool(ABC):
    @abstractmethod
    def claim(self) -> str:
        pass

    @abstractmethod
    def yield_network(self, network: str):
        pass

    @property
    @abstractmethod
    def size(self):
        pass


class StaticNetworkPool(NetworkPool):
    def __init__(self, networks: List[str]):
        self.available_networks = list(networks)
        self.claimed_networks = []

    def claim(self) -> str:
        if not self.available_networks:
            raise ValueError("No network available")

        network = self.available_networks.pop()
        self.claimed_networks.append(network)

        return network

    def yield_network(self, network: str):
        if network not in self.claimed_networks:
            raise ValueError("Trying to yield unclaimed network")
        self.claimed_networks.remove(network)
        self.available_networks.append(network)

    @property
    def size(self):
        return len(self.available_networks) + len(self.claimed_networks)


class DynamicNetworkPool(NetworkPool):
    def __init__(self, networks: List[Tuple[str, int]], builder: Callable):
        self.available_networks = {n[0]: Network(n[0], n[1]) for n in networks}  # type: Dict
        self.claimed_networks = {}
        self.builder = builder

    def claim(self) -> str:
        if not self.available_networks:
            raise ValueError("No network available")

        # Claim one network from the available networks
        network = self.available_networks.pop(list(self.available_networks.keys())[0])

        # Put it in the claimed networks
        self.claimed_networks[network.name] = network
        network.provider = self.builder(network.port)

        # Spin up the dynamic network
        try:
            network.provider.start()
        except ValueError:
            raise

        return network.name

    def yield_network(self, network: str):
        if network not in self.claimed_networks:
            raise ValueError("Network not claimed")

        network = self.claimed_networks.pop(network)

        # Spin down the dynamic network
        try:
            network.provider.stop()
        except ValueError:
            raise

        # Yield the network back
        self.available_networks[network.name] = network


    @property
    def size(self):
        return len(self.available_networks) + len(self.claimed_networks)


from jsonpath_rw import parse
from json import loads
from pathlib import Path
from eth_vertigo.mutator.source_file import SourceFile


def _get_ast(json_file):
    return loads(json_file.read_text("utf-8"))


def _get_src(src_str: str):
    return [int(e) for e in src_str.split(":")]


def _get_binaryop_info(node: dict):
    """
    Gets info on the binary operation from an ast node
    This ast node must be referencing an binary operation

    :param node: ast node to look for
    :return: the operator, src for the operator
    """
    if node["nodeType"] != "BinaryOperation":
        raise ValueError("Passed node is not a binary operation")

    c_src = _get_src(node["src"])

    original_operator = node["operator"]
    op0_src = _get_src(node["leftExpression"]["src"])
    op1_src = _get_src(node["rightExpression"]["src"])

    if not (c_src[2] == op0_src[2] == op1_src[2]):
        raise ValueError("src fields are inconsistent")

    start = op0_src[0] + op0_src[1]
    length = op1_src[0] - start
    op_src = (start, length, c_src[2])

    return original_operator, op_src


def _get_op_info(node: dict):
    c_src = _get_src(node["src"])

    original_operator = node["operator"]
    op0_src = _get_src(node["leftHandSide"]["src"])
    op1_src = _get_src(node["rightHandSide"]["src"])

    if not (c_src[2] == op0_src[2] == op1_src[2]):
        raise ValueError("src fields are inconsistent")

    start = op0_src[0] + op0_src[1]
    length = op1_src[0] - start
    op_src = (start, length, c_src[2])

    return original_operator, op_src


class SolidityFile(SourceFile):
    def __init__(self, json_path: Path):
        self.json = _get_ast(json_path)
        self.ast = self.json["ast"]
        file = Path(self.json["sourcePath"])
        super().__init__(file)

    def get_binary_op_locations(self):
        path_expr = parse('*..nodeType.`parent`')
        for match in path_expr.find(self.ast):
            if match.value["nodeType"] != "BinaryOperation":
                continue
            yield _get_binaryop_info(match.value)

    def get_if_statement_binary_ops(self):
        path_expr = parse('*..nodeType.`parent`')
        for match in path_expr.find(self.ast):
            if match.value["nodeType"] != "IfStatement":
                continue
            condition = match.value["children"][0]
            yield _get_binaryop_info(condition)

    def get_assignments(self):
        path_expr = parse('*..nodeType.`parent`')
        for match in path_expr.find(self.ast):
            if match.value["nodeType"] != "Assignment":
                continue
            yield _get_op_info(match.value)

    def get_void_calls(self):
        path_expr = parse('*..nodeType.`parent`')
        for match in path_expr.find(self.ast):
            if match.value["nodeType"] != "FunctionCall":
                continue
            function_identifier = match.value["expression"]

            function_typedef = function_identifier["typeDescriptions"]["typeString"]
            if "returns" in function_typedef:
                continue
            if "function" not in function_typedef:
                continue
            if function_identifier["typeDescriptions"]["typeIdentifier"].startswith("t_function_event"):
                continue

            try:
                if "require" in function_identifier["name"]:
                    continue
            except KeyError:
                continue
            yield (None, _get_src(match.value["src"]))

    def get_modifier_invocations(self):
        path_expr = parse('*..nodeType.`parent`')
        for match in path_expr.find(self.ast):
            if match.value["nodeType"] != "ModifierInvocation":
                continue
            yield (None, _get_src(match.value["src"]))



import pathlib
from setuptools import setup, find_packages

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

REQUIREMENTS = (HERE / "requirements.txt")

requirements = [x for x in map(str.strip, REQUIREMENTS.read_text().split("\n")) if x != ""]

setup(
    name="eth-vertigo",
    version="1.2.0",
    description="Mutation Testing for Ethereum Smart Contracts",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/joranhonig/vertigo",
    author="Joran Honig",
    author_email="joran.honig@gmail.com",
    license="GPLv3",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    packages=find_packages(),
    include_package_data=True,
    install_requires=requirements,
    entry_points={
        "console_scripts": [
            "vertigo=eth_vertigo.cli.main:cli",
        ]
    },
)



from eth_vertigo.mutator.source_file import SourceFile
from pathlib import Path
from eth_vertigo.mutator.mutator import Mutator


class SourceMutator(Mutator):
    def mutate(self, source_file: SourceFile, project_directory: Path):
        if not source_file.file.exists():
            return []

        result = []

        result += list(self.mutate_boundary_conditionals(source_file, project_directory))
        result += list(self.mutate_negated_conditionals(source_file, project_directory))
        result += list(self.mutate_math_ops(source_file, project_directory))
        result += list(self.mutate_increments(source_file, project_directory))
        result += list(self.mutate_voids(source_file, project_directory))
        result += list(self.mutate_modifier(source_file, project_directory))

        return result

    @staticmethod
    def mutate_boundary_conditionals(source_file: SourceFile, project_directory: Path):
        raise NotImplementedError

    @staticmethod
    def mutate_negated_conditionals(source_file: SourceFile, project_directory: Path):
        raise NotImplementedError

    @staticmethod
    def mutate_math_ops(source_file: SourceFile, project_directory: Path):
        raise NotImplementedError

    @staticmethod
    def mutate_increments(source_file: SourceFile, project_directory: Path):
        raise NotImplementedError

    @staticmethod
    def mutate_voids(source_file: SourceFile, project_directory: Path):
        raise NotImplementedError

    @staticmethod
    def mutate_modifier(source_file: SourceFile, project_directory: Path):
        raise NotImplementedError



from jsonpath_rw import parse
from subprocess import Popen, PIPE
import re
from json import loads, JSONDecodeError
from pathlib import Path
from eth_vertigo.mutator.source_file import SourceFile


def _get_ast(file: Path):
    # Assert precondition
    if not file.exists():
        raise ValueError("File does not exist")

    # Execute solc
    command = ["/home/walker/installers/solc", "--ast-json"]
    command += [str(file)]
    proc = Popen(command, stdin=PIPE, stdout=PIPE, stderr=PIPE)
    output, err = proc.communicate()

    if output == "":
        raise Exception("Error while retrieving the ast: {}".format(err))

    # Find the different ast's
    found_asts = {}
    current_contract = None
    current_json_lines = []

    for line in output.decode('utf-8').split("\n"):
        if re.match("======= .* =======", line):
            if current_json_lines and current_contract:
                json_string = "\n".join(current_json_lines)
                try:
                    found_asts[current_contract] = loads(json_string)
                except JSONDecodeError:
                    pass

            current_contract = line.split(" ")[1]
            current_json_lines = []
        else:
            current_json_lines.append(line)

    if str(file) in found_asts.keys():
        return found_asts[str(file)]
    return None


def _get_src(src_str: str):
    return [int(e) for e in src_str.split(":")]


def _get_binaryop_info(node: dict):
    """
    Gets info on the binary operation from an ast node
    This ast node must be referencing an binary operation

    :param node: ast node to look for
    :return: the operator, src for the operator
    """
    if node["name"] != "BinaryOperation":
        raise ValueError("Passed node is not a binary operation")

    c_src = _get_src(node["src"])

    original_operator = node["attributes"]["operator"]
    op0_src = _get_src(node["children"][0]["src"])
    op1_src = _get_src(node["children"][1]["src"])

    if not (c_src[2] == op0_src[2] == op1_src[2]):
        raise ValueError("src fields are inconsistent")

    start = op0_src[0] + op0_src[1]
    length = op1_src[0] - start
    op_src = (start, length, c_src[2])

    return original_operator, op_src


def _get_op_info(node: dict):
    c_src = _get_src(node["src"])

    original_operator = node["attributes"]["operator"]
    op0_src = _get_src(node["children"][0]["src"])
    op1_src = _get_src(node["children"][1]["src"])

    if not (c_src[2] == op0_src[2] == op1_src[2]):
        raise ValueError("src fields are inconsistent")

    start = op0_src[0] + op0_src[1]
    length = op1_src[0] - start
    op_src = (start, length, c_src[2])

    return original_operator, op_src


class SolidityFile(SourceFile):
    def __init__(self, file: Path):
        super().__init__(file)
        self.ast = _get_ast(file)

    def get_binary_op_locations(self):
        path_expr = parse('*..name.`parent`')
        for match in path_expr.find(self.ast):
            if match.value["name"] != "BinaryOperation":
                continue
            yield _get_binaryop_info(match.value)

    def get_if_statement_binary_ops(self):
        path_expr = parse('*..name.`parent`')
        for match in path_expr.find(self.ast):
            if match.value["name"] != "IfStatement":
                continue
            condition = match.value["children"][0]
            yield _get_binaryop_info(condition)

    def get_assignments(self):
        path_expr = parse('*..name.`parent`')
        for match in path_expr.find(self.ast):
            if match.value["name"] != "Assignment":
                continue
            yield _get_op_info(match.value)

    def get_void_calls(self):
        path_expr = parse('*..name.`parent`')
        for match in path_expr.find(self.ast):
            if match.value["name"] != "FunctionCall":
                continue
            function_identifier = match.value["childen"][0]
            function_typedef = function_identifier["attributes"]["type"]
            if "returns" in function_typedef:
                continue
            if "function" not in function_typedef:
                continue
            if "require" in function_identifier["attributes"]["value"]:
                continue
            yield (None, _get_src(match.value["src"]))

    def get_modifier_invocations(self):
        return []



from abc import ABC, abstractmethod
from typing import Tuple, List
from eth_vertigo.core.mutation import Mutation


class TestSuggester(ABC):
    """ TestSuggester interface definition

    A TestSuggester is used to provide suggestions for which tests to run to the mutation testing campaign.
    """

    @property
    @abstractmethod
    def is_strict(self) -> bool:
        """ Returns whether this suggester provides strict suggestions

        A strict suggestion is one where only the suggested tests should be ran.

        A non-strict suggestions is one where the suggested tests should be ran first, before executing
        the remainder of the test suite.
        """
        pass

    @abstractmethod
    def suggest_tests(self, mutation: Mutation) -> List:
        """ Request list of test from the suggester as to which tests should be ran on the mutation

        :param mutation: The subject mutation

        :return List of tests to run
        """
        pass
